@@ -1,11 +1,23 @@
-from typing import Any, Dict, List
+from typing import Any, Dict, List, NamedTuple
 import logging
+import time
+import datetime
 
 import aiobotocore.session
+import newrelic.agent
+
 
 logger = logging.getLogger(__name__)
 
 
+class FeatureRetrievalMeta(NamedTuple):
+    cache_misses: int
+    retrieval_ms: float
+    success: bool
+    cache_delay_minutes: float
+
+
+@newrelic.agent.function_trace()
 async def async_batch_get(
     dynamo_client, table_name: str, keys: List[Dict[str, Any]]
 ) -> List[Dict[str, Any]]:
@@ -53,6 +65,7 @@ async def async_batch_get(
     return all_items
 
 
+@newrelic.agent.function_trace()
 def parse_dynamo_item(item: Dict[str, Any]) -> Dict[str, Any]:
     """Parse a DynamoDB attribute map (low-level) to Python types."""
     out: Dict[str, Any] = {}
@@ -76,6 +89,7 @@ def parse_dynamo_item(item: Dict[str, Any]) -> Dict[str, Any]:
     return out
 
 
+@newrelic.agent.function_trace()
 async def set_stream_features(
     *,
     streams: List[Dict[str, Any]],
@@ -85,12 +99,20 @@ async def set_stream_features(
     stream_pk_prefix: str,
     cache_sep: str,
     aio_session: aiobotocore.session.Session | None = None,
-) -> None:
+) -> FeatureRetrievalMeta:
+    time_start = time.perf_counter_ns()
     """Fetch missing features for streams from DynamoDB and fill them into streams."""
     if not streams or not stream_features:
-        return
+        return FeatureRetrievalMeta(
+            cache_misses=0,
+            retrieval_ms=(time.perf_counter_ns() - time_start) * 1e-6,
+            success=True,
+            cache_delay_minutes=0,
+        )
 
     cache_miss: Dict[str, Dict[str, Any]] = {}
+    cache_delay_obj: dict[str, float] = {f: 0 for f in stream_features}
+    now = datetime.datetime.utcnow()
     for f in stream_features:
         for s in streams:
             key = f"{s['streamUrl']}{cache_sep}{f}"
@@ -99,13 +121,23 @@ async def set_stream_features(
                 cached = features_cache.get(key)
                 if cached["value"] is not None:
                     s[f] = cached["value"]
+                    cache_delay_obj[f] = max(
+                        cache_delay_obj[f], (now - cached["updated_at"]).total_seconds()
+                    )
             else:
                 cache_miss[key] = s
+    valid_cache_delays = list(v for v in cache_delay_obj.values() if v > 0)
+    cache_delay = min(valid_cache_delays) if valid_cache_delays else 0
 
     if not cache_miss:
-        return
-
-    logger.info("Cache miss for %d items", len(cache_miss))
+        return FeatureRetrievalMeta(
+            cache_misses=0,
+            retrieval_ms=(time.perf_counter_ns() - time_start) * 1e-6,
+            success=True,
+            cache_delay_minutes=cache_delay / 60,
+        )
+    cache_misses = len(cache_miss)
+    logger.info("Cache miss for %d items", cache_misses)
 
     # Prepare keys
     keys = []
@@ -120,7 +152,12 @@ async def set_stream_features(
             items = await async_batch_get(dynamodb, features_table, keys)
         except Exception as e:
             logger.error("DynamoDB batch_get failed: %s", e)
-            return
+            return FeatureRetrievalMeta(
+                cache_misses=cache_misses,
+                retrieval_ms=(time.perf_counter_ns() - time_start) * 1e-6,
+                success=False,
+                cache_delay_minutes=cache_delay / 60,
+            )
 
     updated_keys = set()
     for item in items:
@@ -132,6 +169,9 @@ async def set_stream_features(
         features_cache[cache_key] = {
             "value": parsed.get("value"),
             "cache_ttl_in_seconds": int(parsed.get("cache_ttl_in_seconds", -1)),
+            "updated_at": datetime.datetime.fromisoformat(
+                parsed.get("updated_at")
+            ).replace(tzinfo=None),
         }
         if cache_key in cache_miss:
             cache_miss[cache_key][feature_name] = parsed.get("value")
@@ -142,3 +182,9 @@ async def set_stream_features(
         missing_keys = set(cache_miss.keys()) - updated_keys
         for k in missing_keys:
             features_cache[k] = {"value": None, "cache_ttl_in_seconds": 300}
+    return FeatureRetrievalMeta(
+        cache_misses=cache_misses,
+        retrieval_ms=(time.perf_counter_ns() - time_start) * 1e-6,
+        success=True,
+        cache_delay_minutes=cache_delay / 60,
+    )
